;################################################################################
;#										#
;# chipbasic - single chip basic computer with ATmega32				#
;# runtime engine								#
;# copyright (c) 2006/2007 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 2		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

tbrun:
;-----------------------------------------------------------------------
; clear all variables
;-----------------------------------------------------------------------
		ldi	ZH,HIGH(varspace)	
		ldi	ZL,LOW(varspace)
		ldi	XL,0			;zero
		ldi	XH,180			;variables
tbrun_i1:	st	Z+,XL
		dec	XH
		brne	tbrun_i1
		ldi	ZH,HIGH(bas_ram)	
		ldi	ZL,LOW(bas_ram)	
		ldi	XH,16			;basic system
tbrun_i2:	st	Z+,XL
		dec	XH
		brne	tbrun_i2
		
		ldi	ZH,HIGH(bas_array)	
		ldi	ZL,LOW(bas_array)	
		ldi	XH,0			;array
tbrun_i3:	st	Z+,XL
		dec	XH
		brne	tbrun_i3


		clr	r13			;statement 1
		clr	r12			;line 1
		ldi	XL,7			;white color
		sts	libmio_color,XL		;set color
		ldi	XL,0xc6
		out	ADCSR,XL		;enable ADC
		lds	tempreg1,libmio_ram+46	;program number
		sts	libmio_ram+47,tempreg1	;temp program number (run)
		call	mem_setprog		;set program number
		
;-----------------------------------------------------------------------
; main loop for all lines
;-----------------------------------------------------------------------
tbrun_02:	mov	ctrl,r12		;set line number
		call	mem_readline		;read line from EEPROM
		rcall	basrun			;interpret line
		cpi	ereg,0x00		
		brne	tbrun_03
		mov	XL,r12			;linenumber
		cpi	XL,bas_lines
		brcs	tbrun_02
tbrun_03:	ldi	ctrl,0x0c		;standard format
		ret				;line numer exceeds 0x0f		

tbrun_mtest:	lds	XL,libmio_kflags	;break
		andi	XL,4
		cpi	XL,4			;monitor
		brne	tbrun_mtest1
		call	mon_main		;call monitor
tbrun_mtest1:	rjmp	basrun_05		;end
	

;-----------------------------------------------------------------------
; interpret line from Z
;-----------------------------------------------------------------------
basrun:		ldi	ZH,HIGH(bas_linebuf)	;start of buffer
		ldi	ZL,LOW(bas_linebuf)	;start of buffer
		sts	libmio_channel,byte0	;set output to screen
		clr	ereg			;clear Error
		mov	XL,r13			;statement to go
		clr	XH
basrun_01:	cp	XL,XH			;found?
		breq	basrun_05		;yes

basrun_02:	ld	tempreg4,Z+		;get char
		cpi	tempreg4,0xff		;end of line
		breq	basrun_07		;next line
		cpi	tempreg4,0x27		;comment
		breq	basrun_07		;next line
		cpi	tempreg4,'"'		;string
		brne	basrun_04
basrun_03:	ld	tempreg4,Z+		;get char
		cpi	tempreg4,0xff		;end of line
		breq	basrun_07		;yes
		cpi	tempreg4,0x27		;comment
		breq	basrun_07	
		cpi	tempreg4,'"'		;string
		brne	basrun_03
		rjmp	basrun_02
basrun_04:	cpi	tempreg4,':'		;separator	
		brne	basrun_02
		inc	XH			;statement+1
		cp	XL,XH			;found?
		brne	basrun_02
		rjmp	basrun_05	

basrun_05:	lds	tempreg4,libmio_kflags	;break
		sbrs	tempreg4,0
		rjmp	basrun_06
		ldi	ereg,1			;set ereg to 1
		andi	tempreg4,0xfe		;clear break flag
		sts	libmio_kflags,tempreg4
		ret				;end

basrun_06:	libmio_screenshot
		ld	tempreg4,Z+		;get char
		cpi	tempreg4,0xff		;EOL?
		brne	basrun_08
basrun_07:	clr	r13			;first statement
		inc	r12			;next line
		ret				;end of line

basrun_08:	sbrc	tempreg4,7
		rjmp	basrun_kw		;token
		cpi	tempreg4,32		;space
		breq	basrun_05
		cpi	tempreg4,':'
		brne	basrun_10
		inc	r13
		rjmp	basrun_05
basrun_10:	cpi	tempreg4,'?'		;print
		brne	basrun_10a
		rjmp	bas_print_01	

basrun_10a:	cpi	tempreg4,'@'		;print
		brne	basrun_11
		rjmp	basrun_ar	
		
basrun_11:	cpi	tempreg4,0x27		;comment
		breq	basrun_07		;next line
		cpi	tempreg4,'A'
		brcs	basrun_se
		cpi	tempreg4,'Z'+1
		brcc	basrun_se
		mov	tempreg7,tempreg4	;store variable number
		
;-----------------------------------------------------------------------
; check for variable setting
;-----------------------------------------------------------------------
		rcall	bas_ispace    	;space
	    	cpi	tempreg4,'='
		brne	basrun_se
		rcall	expar			;call parser
		cpi	ereg,0
		brne	basrun_end
		mov	tempreg4,tempreg7
		subi	tempreg4,'A'
		lsl	tempreg4			
		ldi	YH,high(varspace)	
		ldi	YL,low(varspace)
		add	YL,tempreg4
		adc	YH,byte0					
		st	Y+,XL
		st	Y+,XH
		rjmp	tbrun_mtest		;monitor test

basrun_arse:	pop	ZL			;restore pointer
		pop	ZH
basrun_se:	ldi	ereg,7			;syntax error
basrun_end:	ret		

;-----------------------------------------------------------------------
; check for array setting
;-----------------------------------------------------------------------
basrun_ar:	push	ZL			;save pointer
		push	ZH
		ldi	XH,1			;paranthesis pointer	
basrun_ar1:	ld	XL,Z+
		cpi	XL,0xf			;end
		breq	basrun_arse		;syntax error
		cpi	XL,40			;"("
		brne	basrun_ar1
basrun_ar2:	ld	XL,Z+
		cpi	XL,0xff			;end
		breq	basrun_arse		;syntax error
		cpi	XL,40			;"("
		brne	basrun_ar3
		inc	XH			;paranthesis counter +1
		rjmp	basrun_ar2
basrun_ar3:	cpi	XL,41			;")"
		brne	basrun_ar2
		subi	XH,1			;paranthesis counter -1
		brne	basrun_ar2		;at the end?
basrun_ar4:	ld	XL,Z+
		cpi	XL,32			;space
		breq	basrun_ar4
		cpi	XL,'='			;must be an equation
		brne	basrun_arse
		ldi	XL,','			;separator			
		st	-Z,XL			;replace equal with this
		pop	ZH			;restore pointer
		pop	ZL
		clr	ereg
		rcall	expar			;get arrayindex
		sts	bas_ram+10,XL		;store L
		sts	bas_ram+11,XH
basrun_ar5:	ld	XL,Z+
		cpi	XL,32			;space
		breq	basrun_ar5
		cpi	XL,','
		brne	basrun_se		;error
		rcall	expar			;call parser
		cpi	ereg,0
		brne	basrun_end
		rcall	bas_warr		;write array value
		rjmp	tbrun_mtest		;monitor test

;-----------------------------------------------------------------------
; check for keywords
;-----------------------------------------------------------------------
basrun_kw:	mov	r14,ZL			;store pointer
		mov	r15,ZH
		ldi	ZL,LOW(basrun_kw1)
		ldi	ZH,HIGH(basrun_kw1)
		andi	tempreg4,0x3f		;64 token
		add	ZL,tempreg4
		adc	ZH,byte0
		ijmp
		
basrun_kw1:	rjmp	bas_print		;80
		rjmp	bas_cls			;81
		rjmp	bas_goto		;82
		rjmp	bas_box			;83
		rjmp	bas_fbox		;84
		rjmp	bas_sync		;85
		rjmp	bas_end			;86
		rjmp	bas_if			;87
		rjmp	bas_then		;88
		rjmp	bas_posxy		;89
		rjmp	bas_note		;8a
		rjmp	bas_emit		;8b
		rjmp	bas_waitkey		;8c
		rjmp	bas_readkey		;8d
		rjmp	bas_call		;8e
		rjmp	bas_return		;8f
		
		rjmp	bas_for			;90
		rjmp	bas_next		;91
		rjmp	bas_input		;92
		rjmp	bas_plot		;93
		rjmp	bas_wait		;94
		rjmp	bas_tset		;95
		rjmp	bas_tget		;96
		rjmp	bas_eeput		;97
		rjmp	bas_eeget		;98
		rjmp	bas_sput		;99
		rjmp	bas_sget		;9a
		rjmp	bas_epoke		;9b
		rjmp	bas_dir			;9c
		rjmp	bas_out			;9d
		rjmp	bas_color		;9e
		rjmp	bas_lchar		;9f
		
		rjmp	bas_lpix		;a0
		rjmp	bas_chpump		;a1
		rjmp	bas_xpoke		;a2
		rjmp	bas_emit		;a3 +YEmit
		rjmp	bas_limit		;a4
		rjmp	bas_gchar		;a5
		rjmp	bas_draw		;a6
		rjmp	bas_sprite		;a7
		rjmp	bas_show		;a8
		rjmp	bas_hide		;a9
		rjmp	bas_breakpoint		;aa
		rjmp	bas_fast		;ab
		rjmp	bas_slow		;ac
		rjmp	bas_xcall		;ad
		rjmp	bas_icomm		;ae
		rjmp	bas_data		;af
		
		rjmp	bas_ctext		;b0
		rjmp	bas_fread		;b1
		rjmp	bas_fwrite		;b2
		rjmp	bas_aprint		;b3
		rjmp	basrun_kw2		;b4 unknown
		rjmp	basrun_kw2		;b5 unknown
		rjmp	basrun_kw2		;b6 unknown
		rjmp	basrun_kw2		;b7 unknown
		rjmp	basrun_kw2		;b8 unknown
		rjmp	basrun_kw2		;b9 unknown
		rjmp	basrun_kw2		;ba unknown
		rjmp	basrun_kw2		;bb unknown
		rjmp	basrun_kw2		;bc unknown
		rjmp	basrun_kw2		;bd unknown
		rjmp	basrun_kw2		;be unknown
basrun_kw2:	ldi	ereg,8			;unknown
		ret
				
;-----------------------------------------------------------------------
; keyword is PRINT
;-----------------------------------------------------------------------
bas_print:	rcall	bas_rlp			;restore Z-register
bas_print_01:	ldi	ctrl,0x0c		;standard format
		mov	tempreg8,byte1		;enable newline
		sts	libmio_channel,byte0	;print to screen
bas_print_02:	ld	XH,Z+
		cpi	XH,32			;space
		breq	bas_print_02
		cpi	XH,':'
		breq	bas_print_07
		cpi	XH,0x3b			;";"
		breq	bas_print_04
		cpi	XH,','			;tab
		brne	bas_print_05
		rcall	bas_tab
bas_print_04:	clr	tempreg8		;no newline
		rjmp	bas_print_02
bas_print_05:	cpi	XH,0xff
		brne	bas_print_09		;no end
bas_print_07:	sbiw	ZL,1			;points to this char		
		sbrc	tempreg8,0		;newline?
		libmio_newline
bas_print_08:	rjmp	tbrun_mtest		;return to main loop

bas_print_09:	mov	tempreg8,byte1        	;enable newline

		cpi	XH,'"'			;is it a string?
		brne	bas_print_11		;no
bas_print_10:	ld	tempreg1,Z+		;get string char
		cpi	tempreg1,0xff
		breq	bas_print_07		;yes
		cpi	tempreg1,'"'		;end of string
		breq	bas_print_02		;yes
		libmio_outchar			;output char
		rjmp	bas_print_10		;string loop

bas_print_11:	cpi	XH,'!'			;format
		brne	bas_print_11a
		rcall	expar
		mov	ctrl,XL
		rjmp	bas_print_02				

bas_print_11a:	cpi	XH,'%'			;char
		brne	bas_print_12
		rcall	expar
		mov	tempreg1,XL
		libmio_outchar
		rjmp	bas_print_02				

bas_print_12:	cpi	XH,'#'			;format
		brne	bas_print_13
		ld	tempreg1,Z+		;get string char
		andi	tempreg1,0x03		;mask bits
		sts	libmio_channel,tempreg1
		rjmp	bas_print_02				

bas_print_13:	sbiw	ZL,1			;set pointer to char before
		rcall	expar			;mathematics
		cpi	ereg,0
		breq	bas_print_14
		ret
		
bas_print_14:	cpi	ctrl,0x80		;Hex/dez
		brcc	bas_print_15
		libmio_outdez		
		rjmp	bas_print_02				

bas_print_15:	libmio_outhex
		rjmp	bas_print_02				

;-----------------------------------------------------------------------
; keyword is APRINT
;-----------------------------------------------------------------------
bas_aprint_err:	ldi	ereg,7			;syntax error
		ret
bas_aprint:	rcall	exparser	
		sts	bas_ram+10,XL		;array index
		sts	bas_ram+11,XH
		rcall	bas_ispc		;wait for no space
		cpi	tempreg4,','		;comma?
		brne	bas_aprint_err		;no -> end
		adiw	ZL,1
	
bas_aprint_01:	ldi	ctrl,0x03		;channel set
		sts	libmio_channel,ctrl	;print to array
		ldi	ctrl,0x0c		;standard format
		mov	tempreg8,byte1		;enable newline
bas_aprint_02:	ld	XH,Z+
		cpi	XH,32			;space
		breq	bas_aprint_02
		cpi	XH,':'			;next statement
		breq	bas_aprint_03
		cpi	XH,0xff			;EOL
		breq	bas_aprint_03
		cpi	XH,0x3b			;";"
		breq	bas_aprint_02
		cpi	XH,','			;tab
		brne	bas_aprint_04
bas_ap_tab:	lds	XL,bas_ram+10		;arraypos
		andi	XL,0x07			;mask bits
		breq	bas_ap_tab1
		ldi	XL,32			;space
		clr	XH
		rcall	bas_warr		;write to array
		rjmp	bas_ap_tab
bas_ap_tab1:	rjmp	bas_aprint_02		;loop
			
bas_aprint_03:	clr	XL			;end with 0x00
		clr	XH
		rcall	bas_warr
		sbiw	ZL,1			;points to this char		
		rjmp	tbrun_mtest		;return to main loop

bas_aprint_04:	cpi	XH,'"'			;is it a string?
		brne	bas_aprint_11		;no
bas_aprint_10:	ld	XL,Z+			;get string char
		cpi	XL,0xff
		breq	bas_aprint_err		;yes
		cpi	XL,'"'			;end of string
		breq	bas_aprint_02		;yes
		clr	XH
		rcall	bas_warr		;write array value
		rjmp	bas_aprint_10		;string loop

bas_aprint_11:	cpi	XH,'!'			;format
		brne	bas_aprint_12
		rcall	expar
		mov	ctrl,XL
		rjmp	bas_aprint_02				

bas_aprint_12:	cpi	XH,'%'			;char
		brne	bas_aprint_13
		rcall	expar
		mov	tempreg1,XL
		libmio_outchar
		rjmp	bas_aprint_02				

bas_aprint_13:	sbiw	ZL,1			;set pointer to char before
		rcall	expar			;mathematics
		cpi	ereg,0
		breq	bas_aprint_14
		ret
		
bas_aprint_14:	cpi	ctrl,0x80		;Hex/dez
		brcc	bas_aprint_15
		libmio_outdez		
		rjmp	bas_aprint_02				

bas_aprint_15:	libmio_outhex
		rjmp	bas_aprint_02				

;-----------------------------------------------------------------------
; keyword is CLS
;-----------------------------------------------------------------------
bas_cls:	rcall	bas_rlp			;restore Z-register
		libmio_clrscr
		rjmp	tbrun_mtest
	    
;-----------------------------------------------------------------------
; keyword is GOTO
;-----------------------------------------------------------------------
bas_goto:	rcall	exparser
		subi	XL,1
		mov	r12,XL		
		clr	r13
		ret		

;-----------------------------------------------------------------------
; keyword is box
;-----------------------------------------------------------------------
bas_box:	rcall	bas_g4
		libmio_box
		rjmp	bas_g4_end		

;-----------------------------------------------------------------------
; keyword is fbox
;-----------------------------------------------------------------------
bas_fbox:	rcall	bas_g4			;get parameters
		libmio_fbox
		rjmp	bas_g4_end		
		
;-----------------------------------------------------------------------
; keyword is sync
;-----------------------------------------------------------------------
bas_sync:	rcall	exparser
		cpi	ereg,0
		breq	bas_sync_01
		ret
bas_sync_01:	cpi	XL,0
		brne	bas_sync_02
		cpi	XH,0
		brne	bas_sync_02
		rjmp	tbrun_mtest		;continue
bas_sync_02:	libmio_sync
		sbiw	XL,1
		rjmp	bas_sync_01			

;-----------------------------------------------------------------------
; keyword is end
;-----------------------------------------------------------------------
bas_end:	pop	XH
		pop	XH
		ret

;-----------------------------------------------------------------------
; keyword is if
;-----------------------------------------------------------------------
bas_if:		rcall	exparser
		or	XL,XH
		brne	bas_if1
		rjmp	basrun_07		;next line
bas_if1:	rjmp	tbrun_mtest		;continue

;-----------------------------------------------------------------------
; keyword is then
;-----------------------------------------------------------------------
bas_then:	rcall	bas_rlp
		rjmp	tbrun_mtest		;continue

;-----------------------------------------------------------------------
; keyword is pos
;-----------------------------------------------------------------------
bas_posxy:	rcall	bas_g2			;get parameters
		libmio_gotoxy
bas_posxy_ext:	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is note
;-----------------------------------------------------------------------
bas_note:	rcall	exparser
		libmio_note
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is emit
;-----------------------------------------------------------------------
bas_emit:	mov	tempreg3,tempreg4
		rcall	exparser
bas_emit_01:	mov	tempreg1,XL
		libmio_outchar
		sbrc	tempreg3,5		;skip if (X)Emit
		libmio_vcur			;move cursor for YEmit
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_posxy_ext
		rcall	bas_ispace
		call	expar
		rjmp	bas_emit_01	

;-----------------------------------------------------------------------
; keyword is wkey 
;-----------------------------------------------------------------------
bas_waitkey:	rcall	bas_getvar		;get variable
		sts	libmio_lastkey,byte0
bas_waitkey_1:	lds	XL,libmio_lastkey
		cpi	XL,0
		breq	bas_waitkey_1
		st	Y+,XL			;low key
		st	Y+,byte0			
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is rkey
;-----------------------------------------------------------------------
bas_readkey:	rcall	bas_getvar		;get variable
		lds	XL,libmio_keycode
		st	Y+,XL			;low key
		st	Y+,byte0			
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is CALL
;-----------------------------------------------------------------------
bas_call:	rcall	exparser
		cpi	XH,00
		breq	bas_call_02
bas_call_01:	ldi	ereg,10			;no valid linenumber
		ret
bas_call_02:	cpi	XL,0x00
		breq	bas_call_01
		cpi	XL,bas_lines+1		;max linenumber
		brcc	bas_call_01
		lds	XH,bas_ram+1		;stackpointer
		cpi	XH,bas_stackdepth-1	;maximum-1
		brcs	bas_call_03
		ldi	ereg,13			;too many gosub
		ret
bas_call_03:	subi	XL,1			;gosub line
		push	XL			
		ldi	YL,LOW(bas_stacks)	;stack root
		ldi	YH,HIGH(bas_stacks)
		ldi	XL,5			;bytes/record
		mul	XL,XH
		add	YL,r0
		adc	YH,r1
		inc	XH			;inc pointer
		sts	bas_ram+1,XH
		mov	XH,r13			;statement
		inc	XH			;must return to next statement
		ori	XH,0x80			;mark as gosub
		st	Y+,XH			;return statement number
		mov	XH,r12			;our line number
		st	Y+,XH
		lds	XH,libmio_ram+46	;prog number
		st	Y+,XH			;store
		clr	r13			;statement=0
		pop	r12
		ret		

;-----------------------------------------------------------------------
; keyword is XCALL
;-----------------------------------------------------------------------
bas_xcall:	rcall	bas_g2			;XH=prog, XL=line
		dec	XH
		andi	XH,0x03
		sts	libmio_ram+47,XH	;set temp prognumber
		cpi	XH,bas_maxprog
		brcs	bas_xcall_02
bas_xcall_01:	ldi	ereg,10			;no valid linenumber
		ret
bas_xcall_02:	cpi	XL,0x00
		breq	bas_xcall_01
		cpi	XL,bas_lines+1		;max linenumber
		brcc	bas_xcall_01
		lds	XH,bas_ram+1		;stackpointer
		cpi	XH,bas_stackdepth-1	;maximum-1
		brcs	bas_xcall_03
		ldi	ereg,13			;too many gosub
		ret
bas_xcall_03:	subi	XL,1			;gosub line
		push	XL
					
		ldi	YL,LOW(bas_stacks)	;stack root
		ldi	YH,HIGH(bas_stacks)
		ldi	XL,5			;bytes/record
		mul	XL,XH
		add	YL,r0
		adc	YH,r1
		inc	XH			;inc pointer
		sts	bas_ram+1,XH
		mov	XH,r13			;statement
		inc	XH			;must return to next statement
		ori	XH,0x80			;mark as gosub
		st	Y+,XH			;return statement number
		mov	XH,r12			;our line number
		st	Y+,XH
		lds	XH,libmio_ram+46	;prog number
		st	Y+,XH			;store
		clr	r13			;statement=0
		pop	r12
		lds	tempreg1,libmio_ram+47	;get temp program
		call	mem_setprog		;set program
		ret		

;-----------------------------------------------------------------------
; keyword is RETURN
;-----------------------------------------------------------------------
bas_return:	rcall	bas_rlp			;restore Z-register
		lds	XH,bas_ram+1		;stackpointer
		cpi	XH,0
		brne	bas_return_02
bas_return_01:	ldi	ereg,12			;return without call
		ret
bas_return_02:	dec	XH
		ldi	YL,LOW(bas_stacks)	;stack root
		ldi	YH,HIGH(bas_stacks)
		ldi	XL,5			;bytes/record
		mul	XL,XH
		add	YL,r0
		adc	YH,r1
		sts	bas_ram+1,XH
		ld	XH,Y+	
		sbrs	XH,7			;check if TOS is call
		rjmp	bas_return_01		;no
		andi	XH,0x7f
		mov	r13,XH			;statement
		ld	r12,Y+			;linenumber
		ld	tempreg1,Y+		;program
		call	mem_setprog
		ret

;-----------------------------------------------------------------------
; keyword is FOR
;-----------------------------------------------------------------------
bas_for_err:	ldi	ereg,7
		ret
bas_for:	rcall	bas_rlp			;restore Z-register
		rcall	bas_ispace		;space
		cpi	tempreg4,'A'
		brcs	bas_for_err
		cpi	tempreg4,'Z'+1
		brcc	bas_for_err
		subi	tempreg4,'A'
		lsl	tempreg4
		mov	tempreg7,tempreg4	;store variable number
		rcall	bas_ispace		;space
	    	cpi	tempreg4,'='
		brne	bas_for_err		;syntax error
		rcall	expar			;call parser
		cpi	ereg,0
		breq	bas_for_1
		ret
bas_for_1:	ldi	YH,high(varspace)	
		ldi	YL,low(varspace)
		add	YL,tempreg7
		adc	YH,byte0					
		st	Y+,XL			;set variable to initial value
		st	Y+,XH
		rcall	bas_ispace
		cpi	tempreg4,'T'
		brne	bas_for_err
		ld	tempreg4,Z+
		cpi	tempreg4,'O'
		brne	bas_for_err
		rcall	expar		;target value
		mov	tempreg5,XL		;save target value
		mov	tempreg6,XH
		cpi	ereg,0
		breq	bas_for_2
		ret
bas_for_2:	lds	XH,bas_ram+1		;stackpointer
		cpi	XH,bas_stackdepth-1	;maximum-1
		brcs	bas_for_3
		ldi	ereg,14			;too many for
		ret
bas_for_3:	ldi	YL,LOW(bas_stacks)	;stack root
		ldi	YH,HIGH(bas_stacks)
		ldi	XL,5			;bytes/record
		mul	XL,XH
		add	YL,r0
		adc	YH,r1
		inc	XH			;inc pointer
		sts	bas_ram+1,XH
		mov	XH,r13
		inc	XH			;next must skip this
		st	Y+,XH			;return statement number
		mov	XH,r12			;our line number
		st	Y+,XH
		st	Y+,tempreg7		;variable offset
		st	Y+,tempreg5		;target value low
		st	Y+,tempreg6		;target value high
		rjmp	tbrun_mtest	

;-----------------------------------------------------------------------
; keyword is NEXT
;-----------------------------------------------------------------------
bas_next:	rcall	bas_rlp			;restore Z-register
		lds	XH,bas_ram+1		;stackpointer
		cpi	XH,0
		brne	bas_next_02
bas_next_01:	ldi	ereg,11			;next without for
		ret
bas_next_02:	dec	XH
		ldi	YL,LOW(bas_stacks)	;stack root
		ldi	YH,HIGH(bas_stacks)
		ldi	XL,5			;bytes/record
		mul	XL,XH
		add	YL,r0
		adc	YH,r1
		ld	XH,Y+			;statement
		sbrc	XH,7			;check if TOS is for
		rjmp	bas_next_01		;no
		andi	XH,0x7f
		sts	bas_ram+2,XH		;save statement temporary
		ld	tempreg6,Y+		;get line number to jump	
		ld	tempreg4,Y+		;variable
		ldi	XH,high(varspace)	
		ldi	XL,low(varspace)
		add	XL,tempreg4
		adc	XH,byte0					
		ld	r16,X+			;variable low
		ld	r17,X+			;variable high
		add	r16,byte1
		adc	r17,byte0
		st	-X,r17
		st	-X,r16
		ld	XL,Y+			;target low
		ld	XH,Y+			;target high
		adiw	XL,1			;+1
		cp	r16,XL
		brne	bas_next_03
		cp	r17,XH
		brne	bas_next_03
		lds	XH,bas_ram+1		;stackpointer
		dec	XH
		sts	bas_ram+1,XH		
		rjmp	tbrun_mtest		;next statement
bas_next_03:	mov	r12,tempreg6
		lds	r13,bas_ram+2
		ret
			
;-----------------------------------------------------------------------
; keyword is INPUT
;-----------------------------------------------------------------------
bas_input:	rcall	bas_rlp			;restore Z-register
		ldi	XH,0xff
		sts	bas_inbuf,XH
bas_input_00:	ld	XH,Z+
		cpi	XH,32			;space
		breq	bas_input_00
		mov	tempreg1,XH
		cpi	XH,0xff
		brne	bas_input_02
bas_input_01:	libmio_newline
		sbiw	ZL,1
		rjmp	tbrun_mtest

bas_input_02:	cpi	XH,':'			;end of statement
		breq	bas_input_01
		cpi	XH,0x3b			;";"	ignore this
		breq	bas_input_00
		cpi	XH,','			;tab
		brne	bas_input_04
bas_input_03:	rcall	bas_tab
		rjmp	bas_input_00

bas_input_04:	cpi	XH,'"'			;is it a string?
		brne	bas_input_06		;no
bas_input_05:	ld	tempreg1,Z+		;get string char
		cpi	tempreg1,0xff
		breq	bas_input_01
		cpi	tempreg1,'"'		;end of string
		breq	bas_input_00		;yes
		libmio_outchar			;output char
		rjmp	bas_input_05		;string loop
		
bas_input_06:	cpi	XH,'A'
		brcs	bas_input_err
		cpi	XH,'Z'+1
		brcc	bas_input_err
		subi	XH,'A'
		lsl	XH
		sts	bas_ram+2,XH		;variable number
		ld	XH,Z
		cpi	XH,'A'
		brcs	bas_input_07
		cpi	XH,'Z'+1
		brcc	bas_input_07
		rjmp	bas_input_err
bas_input_07:	rcall	bas_inval		;get value		
		ldi	YH,high(varspace)	
		ldi	YL,low(varspace)
		lds	r17,bas_ram+2
		add	YL,r17
		adc	YH,byte0					
		st	Y+,XL			;set variable to initial value
		st	Y+,XH
		rjmp	bas_input_00

bas_input_err:	ldi	ereg,7			;syntax error
		ret

;-----------------------------------------------------------------------
; keyword is plot
;-----------------------------------------------------------------------
bas_plot:	rcall	bas_g2			;get parameters
		libmio_plot
bas_plot_end:	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is wait
;-----------------------------------------------------------------------
bas_wait:	rcall	exparser
		cpi	ereg,0
		breq	bas_wait_01
		ret
bas_wait_01:	mov	tempreg1,XL
		or	tempreg1,XH
		breq	bas_plot_end
bas_wait_02:	lds	tempreg1,libmio_time0
		cpi	tempreg1,0x00
		breq	bas_wait_02
bas_wait_03:	lds	tempreg1,libmio_time0
		cpi	tempreg1,0x00
		brne	bas_wait_03
		sbiw	XL,1
		rjmp	bas_wait_01			    	

;-----------------------------------------------------------------------
; keyword is tset
;-----------------------------------------------------------------------
bas_tset:	rcall	exparser			;get value
		sts	libmio_time_l,XL
		sts	libmio_time_h,XH
		rjmp	bas_wait_01			    	

;-----------------------------------------------------------------------
; keyword is tget
;-----------------------------------------------------------------------
bas_tget:	rcall	bas_getvar		;get variable
		lds	XL,libmio_time_l
		st	Y+,XL			;low timer
		lds	XL,libmio_time_h
		st	Y+,XL			;high timer
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is eset
;-----------------------------------------------------------------------
bas_eeput:	rcall	exparser
		rcall	bas_getprog
		mov	tempreg1,XL
		libeep_write
		mov	tempreg1,XH
		libeep_write
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is eget
;-----------------------------------------------------------------------
bas_eeget:	rcall	bas_getprog
		libeep_read
		mov	XL,tempreg1
		libeep_read
		mov	XH,tempreg1
		rcall	bas_getvar		;get variable
		st	Y+,XL			;low value
		st	Y+,XH			;high value
		rjmp	tbrun_mtest

bas_getprog:	ldi	YL,LOW(1000)		;program number
		ldi	YH,HIGH(1000)
		libeep_read
		adiw	YL,1
		add	YL,tempreg1
		add	YL,tempreg1
		ret

;-----------------------------------------------------------------------
; keyword is sput
;-----------------------------------------------------------------------
bas_sput:	rcall	exparser
bas_sput_01:	mov	tempreg1,XL
		libmio_pser
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_sput_ext
		rcall	bas_ispace
		call	expar
		rjmp	bas_sput_01
bas_sput_ext:	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is sget
;-----------------------------------------------------------------------
bas_sget:	libmio_gser
		mov	XL,tempreg1
		clr	XH
	    	rcall	bas_getvar		;get variable
		st	Y+,XL			;low value
		st	Y+,XH			;high value
		rjmp	tbrun_mtest

;-----------------------------------------------------------------------
; keyword is epoke
;-----------------------------------------------------------------------
bas_epoke:	rcall	bas_g2w			;get parameters
		mov	tempreg1,XL
		cpi	YH,HIGH(1000)		;max EEPROM adr
		brcc	bas_epoke1
		cpi	YL,LOW(1000)		;max EEPROM adr
		brcc	bas_epoke1
		libeep_write
bas_epoke1:	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is xpoke
;-----------------------------------------------------------------------
bas_xpoke:	rcall	bas_g2w
		mov	tempreg2,XL
		mov	XL,YL				;set address
		mov	XH,YH
		ldi	tempreg4,1			;data EEPROM
		libi2c_write
		libmio_sync				;wait min 32ms
		libmio_sync
		libmio_sync
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is dir
;-----------------------------------------------------------------------
bas_dir:	rcall	exparser
		out	DDRA,XL
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is out (XH=pin XL=val)
;-----------------------------------------------------------------------
bas_out:	rcall	bas_g2
		andi	XH,0x07			;8 Bits
		inc	XH
		clr	tempreg4		;or mask for reset
		ldi	tempreg3,0x7f		;and mask for reset
		sbrs	XL,0			;value bit
		ldi	tempreg4,0x80		;or mask for set
		sbrs	XL,0			;value bit
		ldi	tempreg3,0xff		;and mask for set
bas_out_1:	bst	tempreg4,7		;rotate or mask
		lsl	tempreg4
		bld	tempreg4,0
		bst	tempreg3,7		;rotate and mask
		lsl	tempreg3
		bld	tempreg3,0
		dec	XH
		brne	bas_out_1
		in	r16,PORTA
		or	r16,tempreg4
		and	r16,tempreg3
		out	PORTA,r16
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is color
;-----------------------------------------------------------------------
bas_color:	rcall	exparser
		andi	XL,0x07
		sts	libmio_color,XL
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is lchar (lines)
;-----------------------------------------------------------------------
bas_lchar:	rcall	exparser
		libmio_lscroll
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is lpix (lines)
;-----------------------------------------------------------------------
bas_lpix:	rcall	exparser
		libmio_lpscroll
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is pump
;-----------------------------------------------------------------------
bas_chpump:	rcall	exparser
		libmio_chpump
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is limit
;-----------------------------------------------------------------------
bas_limit:	rcall	bas_getvar		;variable
		sts	bas_ram+8,YL		;save var address
		sts	bas_ram+9,YH
		ld	r16,Y+			;set var value
		ld	r17,Y+
		subi	r17,0x80		;change sign bit
		rcall	bas_ispace
		cpi	tempreg4,','
		brne	bas_limit_err
		rcall	bas_g2wa		;get min & max value		
		subi	XH,0x80			;change sign bit
		subi	YH,0x80
		cp	YL,r16			;lower limit
		cpc	YH,r17
		brcs	bas_limit1		;is greater
		mov	r16,YL
		mov	r17,YH
bas_limit1:	cp	XL,r16		
		cpc	XH,r17
		brcc	bas_limit2
		mov	r16,XL
		mov	r17,XH
bas_limit2:	lds	YH,bas_ram+9
		lds	YL,bas_ram+8
		subi	r17,0x80
		st	Y+,r16
		st	Y+,r17
	    	rjmp	tbrun_mtest		
bas_limit_err:	ldi	ereg,6			;syntax error
		ret

;-----------------------------------------------------------------------
; keyword is gchar
;-----------------------------------------------------------------------
bas_gchar:	rcall	bas_getvar		;variable
		sts	bas_ram+8,YL		;save var address
		sts	bas_ram+9,YH
		rcall	bas_ispace
		cpi	tempreg4,','
		brne	bas_limit_err
		rcall	bas_g2a			;get min & max value		
		ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	tempreg4,libmio_cols
		mul	tempreg4,XH
		add	YL,r0
		adc	YH,r1
		add	YL,XL
		adc	YH,byte0
		clr	XH
		ld	XL,Y
		lds	YH,bas_ram+9
		lds	YL,bas_ram+8
		st	Y+,XL
		st	Y+,XH
	    	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is draw
;-----------------------------------------------------------------------
bas_draw:	rcall	bas_g4			;get parameters
		libmio_draw
		rjmp	bas_g4_end		

;-----------------------------------------------------------------------
; keyword is sprite
;-----------------------------------------------------------------------
bas_sprite:	rcall	bas_g3			;get number,type and char
		libmio_sdef			;define Sprite
	  	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is show
;-----------------------------------------------------------------------
bas_show:	rcall	bas_g3			;get number, X and Y
		libmio_sshow
	  	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is break
;-----------------------------------------------------------------------
bas_breakpoint:	rcall	bas_rlp
		lds	XL,libmio_kflags
		ori	XL,4			;set monitor flag
		sts	libmio_kflags,XL
	  	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is fast
;-----------------------------------------------------------------------
bas_fast:	lds	XL,libmio_config
		ori	XL,0x20
		sts	libmio_config,XL
	  	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is slow
;-----------------------------------------------------------------------
bas_slow:	lds	XL,libmio_config
		andi	XL,0xdf
		sts	libmio_config,XL
	  	rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is hide
;-----------------------------------------------------------------------
bas_hide:	rcall	exparser		;get number
		mov	ctrl,XL			;number
		libmio_shide
	  	rjmp	tbrun_mtest		


;-----------------------------------------------------------------------
; keyword is icomm
;-----------------------------------------------------------------------
bas_icomm:	rcall	bas_g3			;get adr, start and um
		sts	bas_ram+12,XL		;number of bytes
		libi2c_start			;startbit
		cpi	tempreg3,0x08
		brne	bas_icomm_e		;error
		sbrs	ctrl,0			;skip if read
		rjmp	bas_icomm_w		;jump to write
		mov	tempreg2,ctrl		;address
		libi2c_wbyte			;output
		cpi	tempreg3,0x40
		brne	bas_icomm_e		;error
bas_icomm_r1:	ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		lds	tempreg1,bas_ram+12	;number of bytes
		add	tempreg1,XH		;add offset
		sub	tempreg1,XL		;sub number of bytes to do
		add	YL,tempreg1		;array element address
		adc	YH,byte0
		cpi	XL,1			;last?
		breq	bas_icomm_r2
		libi2c_rbyte			;read byte with ack
		cpi	tempreg3,0x50
		brne	bas_icomm_e		;error
		st	Y,tempreg2		;store byte
		dec	XL			;dec number of byte to read
		rjmp	bas_icomm_r1

bas_icomm_r2:	libi2c_rbyten			;read byte without ack
		cpi	tempreg3,0x58
		brne	bas_icomm_e		;error
		st	Y,tempreg2		;store byte
		libi2c_stop			;
		rjmp	tbrun_mtest		

bas_icomm_w:	mov	tempreg2,ctrl		;address
		libi2c_wbyte			;output
		cpi	tempreg3,0x18
		brne	bas_icomm_e		;error
bas_icomm_w1:	ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		lds	tempreg1,bas_ram+12	;number of bytes
		add	tempreg1,XH		;add offset
		sub	tempreg1,XL		;sub number of bytes to do
		add	YL,tempreg1		;array element address
		adc	YH,byte0
		ld	tempreg2,Y		;get byte
		libi2c_wbyte			;write byte
		cpi	tempreg3,0x28
		brne	bas_icomm_e		;error
		dec	XL			;dec number of byte to read
		brne	bas_icomm_w1
		libi2c_stop
		rjmp	tbrun_mtest		
		
bas_icomm_e:	ldi	ereg,15			;i2c error
		ret				;end

;-----------------------------------------------------------------------
; keyword is DATA
;-----------------------------------------------------------------------
bas_data:	rcall	exparser	
		sts	bas_ram+10,XL		;array index
		sts	bas_ram+11,XH
;Z points to char
		
bas_data_01:	rcall	bas_ispc		;wait for no space
		cpi	tempreg4,','		;comma?
		brne	bas_data_02		;no -> end
		adiw	ZL,1
		rcall	bas_ispace		;next nonspace char
		cpi	tempreg4,'"'		;string?
		breq	bas_data_03		;yes
		sbiw	ZL,1
		call	expar
		rcall	bas_warr		;write array value
		rjmp	bas_data_01	

bas_data_02:	rjmp	tbrun_mtest

bas_data_03:	ld	tempreg4,Z+
		cpi	tempreg4,'"'
		breq	bas_data_04
		cpi	tempreg4,0xff
		breq	bas_data_04
		clr	XH
		mov	XL,tempreg4
		rcall	bas_warr		;write array value
		rjmp	bas_data_03		

bas_data_04:	ld	tempreg4,Z
		rjmp	bas_data_01

;-----------------------------------------------------------------------
; keyword is CTEXT
;-----------------------------------------------------------------------
bas_ctext:	rcall	bas_g2w
		sts	bas_ram+10,YL		;offset
		sts	bas_ram+11,YH		;array area
		lds	tempreg2,bas_ram+12	;number of chars
		cp	XL,tempreg2		;limit
		brcc	bas_ctext_00
		mov	tempreg2,XL
	
bas_ctext_00:	ldi	XL,LOW(bas_inbuf)
		ldi	XH,HIGH(bas_inbuf)
			
bas_ctext_01:	cpi	tempreg2,0
		breq	bas_ctext_02
		ld	tempreg1,X+
		sts	bas_ram+13,XL
		sts	bas_ram+14,XH
		mov	XL,tempreg1
		clr	XH
		rcall	bas_warr
		lds	XL,bas_ram+13
		lds	XH,bas_ram+14		
		dec	tempreg2
		rjmp	bas_ctext_01	
bas_ctext_02:	clr	XL
		clr	XH
		rcall	bas_warr
		rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is FREAD
;-----------------------------------------------------------------------
bas_fread:	rcall	bas_g2w			;2 parameters
		inc	YL			;
		andi	YL,1			
		mov	ctrl,YL
		rcall	dfl_pread
		rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; keyword is FWRITE
;-----------------------------------------------------------------------
bas_fwrite:	rcall	bas_g2w			;2 parameters
		inc	YL			;
		andi	YL,1			
		mov	ctrl,YL
		rcall	dfl_pwrite
		rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; get 4 parameters (YH,YL,ZH,ZL), Z will be saved
;-----------------------------------------------------------------------
bas_g4:		rcall	exparser
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		adiw	ZL,1			;next char
		sts	bas_ram+6,XL		;save X
		rcall	expar
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		adiw	ZL,1			;next char
		sts	bas_ram+7,XL		;save X
		rcall	expar
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		adiw	ZL,1			;next char
		mov	tempreg7,XL		;save X
		rcall	expar
		lds	YH,bas_ram+6
		lds	YL,bas_ram+7
		mov	tempreg5,ZL
		mov	tempreg6,ZH
		mov	ZL,XL
		mov	ZH,tempreg7
		ret

bas_g4_err:	pop	ereg
		pop	ereg
		ldi	ereg,7
		ret

;restore Z-register
bas_g4_end:	mov	ZL,tempreg5
		mov	ZH,tempreg6
		rjmp	tbrun_mtest		

;-----------------------------------------------------------------------
; get 3 parameters (ctrl,XH,XL)
;-----------------------------------------------------------------------
bas_g3:		rcall	exparser
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		mov	ctrl,XL			;save par1
		adiw	ZL,1
		rcall	expar
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		sts	bas_ram+6,XL		;save X
		adiw	ZL,1
		rcall	expar
		lds	XH,bas_ram+6		;par1
		ret

;-----------------------------------------------------------------------
; get 2 byte parameters (XH,XL)
;-----------------------------------------------------------------------
bas_g2a:	call	expar
		rjmp	bas_g2b
bas_g2:		rcall	exparser
bas_g2b:	rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		adiw	ZL,1			;next char
		mov	tempreg7,XL		;save X1
		call	expar
		mov	XH,tempreg7
		ret

;-----------------------------------------------------------------------
; get 2 word parameters (Y,X)
;-----------------------------------------------------------------------
bas_g2wa:	call	expar
		rjmp	bas_g2wb
bas_g2w:	rcall	exparser
bas_g2wb:	sts	bas_ram+6,XL			;save X
		sts	bas_ram+7,XH			;save X
		rcall	bas_ispc
		cpi	tempreg4,','
		brne	bas_g4_err
		adiw	ZL,1			;next char
		call	expar
		lds	YL,bas_ram+6
		lds	YH,bas_ram+7
		ret

;-----------------------------------------------------------------------
; get variable address
;-----------------------------------------------------------------------
bas_getvar:	mov	ZL,r14			;get line pointer
		mov	ZH,r15
bas_va:		rcall	bas_ispace
bas_va_1:	cpi	tempreg4,'A'
		brcs	bas_va_err
		cpi	tempreg4,'Z'+1
		brcc	bas_va_err
		subi	tempreg4,'A'
		lsl	tempreg4
		ldi	YH,high(varspace)	
		ldi	YL,low(varspace)
		add	YL,tempreg4
		adc	YH,byte0					
		ret
bas_va_err:	pop	ereg
		pop	ereg
		ldi	ereg,7
		ret		

;-----------------------------------------------------------------------
; read until char is not space
;-----------------------------------------------------------------------
bas_ispace:	ld	tempreg4,Z+
bas_ispc:	cpi	tempreg4,32			;space
		breq	bas_ispace
bas_ignoresp01:	ret		

;-----------------------------------------------------------------------
; restore line pointer
;-----------------------------------------------------------------------
bas_rlp:	mov	ZL,r14		;get line pointer
		mov	ZH,r15
		ret

;-----------------------------------------------------------------------
; restore line pointer and get parameter
;-----------------------------------------------------------------------
exparser:	mov	ZL,r14		;get line pointer
		mov	ZH,r15
		rjmp	expar

;-----------------------------------------------------------------------
; output tab
;-----------------------------------------------------------------------
bas_tab:	libmio_outspace
		mov	XH,libmio_cur_x
		andi	XH,0x07
		brne	bas_tab
		ret

;-----------------------------------------------------------------------
; input value
;-----------------------------------------------------------------------
bas_inval:	push	ZL			;save pointer
		push	ZH
		libmio_calc			;calculate address
		mov	ZL,YL			;set start of pointer
		mov	ZH,YH
		ldi	r16,0			;length
bas_inval_01:	libmio_cursor			;view cursor
		lds	tempreg4,libmio_kflags	;break
		cpi	tempreg4,1
		brne	bas_inval_01b
		mov	ereg,tempreg4
		rjmp	bas_inval_01a	
	
bas_inval_01b:	cpi	tempreg1,0xea		;enter
		brne	bas_inval_04
		sts	bas_ram+12,r16		;number of chars
		cpi	r16,0			;no char
		brne	bas_inval_02	
		clr	XL
		clr	XH
bas_inval_01a:	clr	ereg
		pop	ZH			;restore pointer
		pop	ZL
		ret
		
bas_inval_02:	ldi	YL,LOW(bas_inbuf)
		ldi	YH,HIGH(bas_inbuf)
bas_inval_03:	ld	XL,Z+
		st	Y+,XL
		dec	r16
		brne	bas_inval_03
		ldi	XL,0x20
		st	Y+,XL
		ldi	XL,0xff			;eol
		st	Y+,XL
		ldi	ZL,LOW(bas_inbuf)
		ldi	ZH,HIGH(bas_inbuf)
		call	expar			;calculate value
		rjmp	bas_inval_01a	
bas_inval_04:	cpi	tempreg1,0xec		;backspace
		brne	bas_inval_05
		cpi	r16,0
		breq	bas_inval_01		;nothing
		dec	libmio_cur_x
		libmio_calc
		ldi	XL,0x20
		st	Y,XL
		dec	r16
		rjmp	bas_inval_01
bas_inval_05:	cpi	tempreg1,0x20
		brcs	bas_inval
		cpi	tempreg1,'z'+1
		brcc	bas_inval
		mov	XL,libmio_cur_x		
		cpi	XL,libmio_cols-1
		brcc	bas_inval
		libmio_outchar
		inc	r16
		rjmp	bas_inval_01
		
bas_vline:	ldi	ctrl,2
		clr	XH
		mov	XL,r12		;line
		inc	XL
		libmio_outdez
		ldi	tempreg1,':'
		libmio_outchar
		clr	XH
		mov	XL,r13		;statement
		libmio_outdez
		libmio_outspace
		ret				

;-----------------------------------------------------------------------
; write array value
;-----------------------------------------------------------------------
bas_warr:	lds	tempreg2,bas_ram+10	;index low
		lds	tempreg1,bas_ram+11	;index high
		sbrs	tempreg1,2		;byte at 1024++
		rjmp	bas_waw			;-> word
		andi	tempreg1,0xfb		;delete word/byte bit

;-----------------------------------------------------------------------
; write array byte (XL to bas_ram+10/11)
;-----------------------------------------------------------------------
bas_wab:	cpi	tempreg1,0x00
		brne	bas_wab_01
		ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		add	YL,tempreg2		;array element address
		adc	YH,byte0
		st	Y+,XL
		rjmp	arr_echeck
		
bas_wab_01:	cpi	tempreg1,0x01
		brne	bas_wab_02
		clr	tempreg3
		rcall	dfl_bw
		rjmp	arr_echeck	

bas_wab_02:	cpi	tempreg1,0x02
		brne	bas_wab_03
		ldi	tempreg3,0x01
		rcall	dfl_bw
		rjmp	arr_echeck	

bas_wab_03:	ldi	ereg,18			;out of array
		rjmp	arr_echeck1
		
;-----------------------------------------------------------------------
; write array word (X to bas_ram+10/11)
;-----------------------------------------------------------------------
bas_waw:	lsl	tempreg2		;*2
		rol	tempreg1
		cpi	tempreg1,0x00
		brne	bas_waw_01
		ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		add	YL,tempreg2		;array element address
		adc	YH,byte0
		st	Y+,XL
		st	Y+,XH
		rjmp	arr_echeck

bas_waw_01:	cpi	tempreg1,0x01
		brne	bas_waw_02
		clr	tempreg3
		rcall	dfl_ww
		rjmp	arr_echeck	

bas_waw_02:	cpi	tempreg1,0x02
		brne	bas_waw_03
		ldi	tempreg3,0x01
		rcall	dfl_ww
		rjmp	arr_echeck	

bas_waw_03:	ldi	ereg,18			;out of array
		rjmp	arr_echeck1
				
;-----------------------------------------------------------------------
; read array value
;-----------------------------------------------------------------------
bas_rarr:	lds	tempreg2,bas_ram+10	;index low
		lds	tempreg1,bas_ram+11	;index high
		sbrs	tempreg1,2		;byte/word
		rjmp	bas_raw			;-> word
		andi	tempreg1,0xfb		;delete word/byte bit

;-----------------------------------------------------------------------
; read array byte (bas_ram+10/11 to XL)
;-----------------------------------------------------------------------
bas_rab:	cpi	tempreg1,0x00
		brne	bas_rab_01
		ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		lds	tempreg1,bas_ram+10	;offset
		add	YL,tempreg1		;array element address
		adc	YH,byte0
		ld	XL,Y+
		clr	XH
		rjmp	arr_echeck

bas_rab_01:	cpi	tempreg1,0x01
		brne	bas_rab_02
		clr	tempreg3
		rcall	dfl_br
		ret

bas_rab_02:	cpi	tempreg1,0x02
		brne	bas_rab_03
		ldi	tempreg3,0x01
		rcall	dfl_br
		ret
			
bas_rab_03:	ldi	ereg,18
		ret

;-----------------------------------------------------------------------
; read array word (bas_ram+10/11 to X)
;-----------------------------------------------------------------------
bas_raw:	lsl	tempreg2		;*2
		rol	tempreg1
		cpi	tempreg1,0x00
		brne	bas_raw_01
		ldi	YH,high(bas_array)	;array
		ldi	YL,low(bas_array)
		add	YL,tempreg2		;array element address
		adc	YH,byte0
		ld	XL,Y+
		ld	XH,Y+
		rjmp	arr_echeck

bas_raw_01:	cpi	tempreg1,0x01
		brne	bas_raw_02
		clr	tempreg3
		rcall	dfl_wr
		ret

bas_raw_02:	cpi	tempreg1,0x02
		brne	bas_raw_03
		ldi	tempreg3,0x01
		rcall	dfl_wr
		ret

bas_raw_03:	ldi	ereg,18
		ret
		
	
arr_echeck:	lds	tempreg1,bas_ram+10	;index low
		lds	tempreg2,bas_ram+11	;index high
		add	tempreg1,byte1		;+1
		adc	tempreg2,byte0
		sts	bas_ram+10,tempreg1	;write back
		sts	bas_ram+11,tempreg2
		cpi	ereg,0
		brne	arr_echeck1
		ret
		
arr_echeck1:	pop	tempreg1
		pop	tempreg1
		ret		